/*******************************************************************************

Copyright (c) 2001-2008, Perforce Software, Inc.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1.  Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

2.  Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL PERFORCE SOFTWARE, INC. BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/


/*******************************************************************************
 * Name		: p4result.cc
 *
 * Author	: Tony Smith <tony@perforce.com> or <tony@smee.org>
 *
 * Description	: Ruby class for holding results of Perforce commands 
 *
 ******************************************************************************/
#include <clientapi.h>
#include "perlheaders.h"
#include "p4perldebug.h"
#include "p4result.h"

P4Result::P4Result()
{
    debug  = 0;
    Init();
}

void
P4Result::Init()
{
    output = newAV();
    errors = newAV();
    warnings = newAV();
}

P4Result::~P4Result()
{
    Clear();
}

void
P4Result::Clear()
{
    sv_free((SV *) output);
    sv_free((SV *) warnings);
    sv_free((SV *) errors);
}

//
// Clean up our output and return it. Here, we decrement the reference 
// count as we're not going to hold on to our output any longer than
// necessary.
//
AV *
P4Result::GetOutput()
{ 
    AV *o = output;
    output = newAV();

    return (AV*) sv_2mortal( (SV*)o );
}

void
P4Result::Reset()
{
    Clear();
    Init();
}

void
P4Result::AddOutput( const char *msg )
{
    if( P4PERL_DEBUG_DATA )
	PerlIO_stdoutf( "[P4Result::AddOutput]: %s\n", msg );

    av_push( output, newSVpv( msg, 0 ) );
}

void
P4Result::AddOutput( SV * out )
{
    if( P4PERL_DEBUG_DATA )
	PerlIO_stdoutf( "[P4Result::AddOutput]: (perl object)\n" );

    av_push( output, out );
}

void
P4Result::AddError( Error *e )
{
    StrBuf	m;
    e->Fmt( &m );

    int s;
    s = e->GetSeverity();

    // 
    // Empty and informational messages are pushed out as output as nothing
    // worthy of error handling has occurred. Warnings go into the warnings
    // list and the rest are lumped together as errors.
    //

    if ( s == E_EMPTY || s == E_INFO )
    {
	AddOutput( m.Text() );
	return;
    }

    if( P4PERL_DEBUG_DATA )
	PerlIO_stdoutf( "[P4Result::AddError]: %s\n", m.Text() );

    if ( s == E_WARN )
	av_push( warnings, newSVpv( m.Text(), 0 ) );
    else
	av_push( errors, newSVpv( m.Text(), 0 ) );
}

I32
P4Result::OutputCount()
{
    return av_len( output ) + 1;
}


I32
P4Result::ErrorCount()
{
    return av_len( errors ) + 1;
}

I32
P4Result::WarningCount()
{
    return av_len( warnings ) + 1;
}

